#include <stdlib.h>
#include <string.h>

#include "lsm.h"
#include "lsm/store_internal.h"

lsm_error lsm_store_load(lsm_store **ptr, lsm_str *data_path) {
  lsm_store *store;
  LSM_RES(lsm_store_init(&store));

  // Try to open an existing db file or create a new one otherwise
  // This shit is why I need to improve the str library
  char db_file_path[lsm_str_len(data_path) + strlen(LSM_DB_FILE_NAME) + 2];
  memcpy(db_file_path, lsm_str_ptr(data_path),
         lsm_str_len(data_path) * sizeof(char));
  sprintf(&db_file_path[lsm_str_len(data_path)], "/%s", LSM_DB_FILE_NAME);

  FILE *db_file = fopen(db_file_path, "r+b");

  if (db_file == NULL) {
    // Create the file first, then reopen it in extended read
    db_file = fopen(db_file_path, "wb");

    if (db_file == NULL) {
      return lsm_error_failed_io;
    }

    fclose(db_file);

    FILE *db_file = fopen(db_file_path, "r+b");

    if (db_file == NULL) {
      return lsm_error_failed_io;
    }
  }

  // Same for idx file
  char idx_file_path[lsm_str_len(data_path) + strlen(LSM_IDX_FILE_NAME) + 2];
  memcpy(idx_file_path, lsm_str_ptr(data_path),
         lsm_str_len(data_path) * sizeof(char));
  sprintf(&idx_file_path[lsm_str_len(data_path)], "/%s", LSM_IDX_FILE_NAME);

  FILE *idx_file = fopen(idx_file_path, "r+b");

  if (idx_file == NULL) {
    // Create the file first
    idx_file = fopen(idx_file_path, "wb");

    if (idx_file == NULL) {
      return lsm_error_failed_io;
    }

    // The database code expects the idx file to start with how many blocks it
    // contains, so we write that here
    uint64_t num = 0;

    if (fwrite(&num, sizeof(uint64_t), 1, idx_file) == 0) {
      return lsm_error_failed_io;
    }

    fflush(idx_file);
    fclose(idx_file);

    // If opening it in extended read mode still fails now, there's a problem
    FILE *idx_file = fopen(idx_file_path, "r+b");

    if (idx_file == NULL) {
      return lsm_error_failed_io;
    }
  }

  store->data_path = data_path;
  store->db_file = db_file;
  store->idx_file = idx_file;

  LSM_RES(lsm_store_load_db(store));

  *ptr = store;

  return lsm_error_ok;
}

static lsm_error lsm_fread(void *out, uint64_t *sum, FILE *f, uint64_t size,
                           uint64_t count) {
  size_t res = fread(out, size, count, f);

  if (res < count) {
    return lsm_error_failed_io;
  }

  if (sum != NULL) {
    *sum += size * count;
  }

  return lsm_error_ok;
}

static lsm_error lsm_entry_read_str(lsm_str **out, uint64_t *sum, FILE *f) {
  uint64_t len;
  LSM_RES(lsm_fread(&len, sum, f, sizeof(uint64_t), 1));

  char *buf = malloc(len + 1);

  if (buf == NULL) {
    return lsm_error_failed_alloc;
  }

  buf[len] = '\0';

  uint64_t read = 0;

  while (read < len) {
    read += fread(&buf[read], 1, len - read, f);
  }

  if (sum != NULL) {
    *sum += len;
  }

  return lsm_str_init(out, buf);
}

static lsm_error lsm_entry_read_attrs(uint64_t *sum, lsm_entry_handle *handle,
                                      FILE *db_file) {
  uint8_t attr_count;
  LSM_RES(lsm_fread(&attr_count, sum, db_file, sizeof(uint8_t), 1));

  // attr_type, val_len
  uint8_t attr_type;
  lsm_str *val;

  for (uint64_t i = 0; i < attr_count; i++) {
    LSM_RES(lsm_fread(&attr_type, sum, db_file, sizeof(uint8_t), 1));
    LSM_RES(lsm_entry_read_str(&val, sum, db_file));
    lsm_entry_attr_insert(handle, attr_type, val);
  }

  return lsm_error_ok;
}

lsm_error lsm_store_load_db(lsm_store *store) {
  uint64_t db_dim[2];
  lsm_str *key;
  lsm_entry_handle *handle;

  rewind(store->idx_file);

  // idx file starts with block count
  LSM_RES(lsm_fread(&store->idx_file_block_count, &store->idx_file_size,
                    store->idx_file, sizeof(uint64_t), 1));

  for (uint64_t i = 0; i < store->idx_file_block_count; i++) {
    LSM_RES(lsm_entry_read_str(&key, &store->idx_file_size, store->idx_file));
    LSM_RES(lsm_fread(&db_dim, &store->idx_file_size, store->idx_file,
                      sizeof(uint64_t), 2));
    LSM_RES(lsm_store_insert(&handle, store, key));

    // Read attributes from database file
    if (fseek(store->db_file, db_dim[0], SEEK_SET) != 0) {
      return lsm_error_failed_io;
    }

    LSM_RES(lsm_fread(&handle->wrapper->entry->data_len, NULL, store->db_file,
                      sizeof(uint64_t), 1));
    LSM_RES(lsm_entry_read_attrs(NULL, handle, store->db_file));
    lsm_entry_close(handle);

    store->db_file_size += db_dim[1];
  }

  return lsm_error_ok;
}
